import json
import os
import pathlib
import uuid

import docker
import jinja2
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.types import OverlayImageBuildException, OverlayImagePushException

TEMPLATES_DIR = f"{pathlib.Path(__file__).parent.resolve()}/templates/"
TEMPLATE_FILE = "Dockerfile.jinja"

KOMODO_DOCKERFILE_NAME = "komodo.Dockerfile"


class OverlayImages:
    def __init__(self, base_image, source_dir, dest_dir, overlay_images_repository):
        self.base_image = base_image
        self.source_dir = source_dir
        self.dest_dir = dest_dir
        self.overlay_images_repository = overlay_images_repository
        self.client = docker.from_env()

    def generate_dockerfile(self) -> str:
        templateLoader = jinja2.FileSystemLoader(searchpath=TEMPLATES_DIR)
        templateEnv = jinja2.Environment(loader=templateLoader)
        template = templateEnv.get_template(TEMPLATE_FILE)
        output = template.render(
            base_image=self.base_image,
            dest_dir=self.dest_dir,
        )
        return output

    def build_overlay_image(self):
        dockerfile = self.generate_dockerfile()
        dockerfile_path = os.path.join(self.source_dir, KOMODO_DOCKERFILE_NAME)
        with open(dockerfile_path, "w") as f:
            f.write(dockerfile)

        try:
            image, build_logs = self.client.images.build(
                path=os.path.dirname(dockerfile_path),
                tag=f"{self.overlay_images_repository}:{str(uuid.uuid4())}",
                quiet=False,
                rm=True,
                pull=True,
                forcerm=True,
                dockerfile=dockerfile_path,
                platform="linux/amd64",  # TODO
            )
            return image
        except (
            docker.errors.BuildError,
            docker.errors.APIError,
            TypeError,
        ) as e:
            raise OverlayImageBuildException(str(e))
        except Exception as e:
            raise e

    def push_overlay_image(self, image):
        try:
            resp = self.client.images.push(
                self.overlay_images_repository,
                image.tags[0].split(":")[-1],
                stream=True,
                decode=True,
            )

            for line in resp:
                if "error" in line:
                    raise OverlayImagePushException(line["error"])
        except Exception as e:
            raise OverlayImagePushException(str(e))



from enum import Enum
from typing import List, Optional


class JobStatus(Enum):
    PENDING = "pending"
    RUNNING = "running"
    FINISHED = "finished"
    CANCELLING = "cancelling"
    CANCELLED = "cancelled"
    ERROR = "error"
    UNKNOWN = "unknown"
    NOT_FOUND = "not found"
    UNAUTHORIZED = "unauthorized"
    UNREACHABLE = "unreachable"
    # this status is for local jobs that were not created on the current computer
    NOT_AVAILABLE_CURRENT_COMPUTER = "not_available_on_current_computer"


LOCAL_STATUS_MAPPING = {
    "cancelling": JobStatus.CANCELLING,
    "cancelled": JobStatus.CANCELLED,
    "created": JobStatus.PENDING,
    "running": JobStatus.RUNNING,
    "paused": JobStatus.RUNNING,
    "restarting": JobStatus.PENDING,
    "removing": JobStatus.FINISHED,
    "exited": JobStatus.FINISHED,
    "dead": JobStatus.FINISHED,
}


class ClientException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class TooManyNodes(ClientException):
    def __init__(self, max_supported_replicas: int):
        super().__init__(
            f"Backend only supports up to {max_supported_replicas} node{'s' if max_supported_replicas > 1 else ''}"
        )


class JobNotFoundException(ClientException):
    def __init__(self):
        super().__init__("Job not found")


class JobNotInDesiredStateException(ClientException):
    def __init__(self):
        super().__init__("Job not in desired state")


class OverlayImageBuildException(ClientException):
    def __init__(self):
        super().__init__("Overlay image failed to build")


class OverlayImagePushException(ClientException):
    def __init__(self, msg):
        super().__init__(
            f"Overlay image failed to push with the following error (please make sure you are logged in with Docker and have the correct permissions to push to your registry): {msg}"
        )


class Job:
    def __init__(
        self,
        backend_name: str,
        resource_name: Optional[str],
        id: str,
        backend_job_id: Optional[str],
        command: str,
        image: str,
        status: JobStatus,
        num_nodes: int,
    ):
        self.backend_name = backend_name
        self.resource_name = resource_name
        self.id = id
        self.backend_job_id = backend_job_id
        self.command = command
        self.image = image
        self.status = status
        self.num_nodes = num_nodes

    @classmethod
    def from_dict(cls, d):
        obj = cls(**d)
        obj.status = JobStatus(d["status"])
        return obj

    def to_dict(self):
        return {
            "backend_name": self.backend_name,
            "resource_name": self.resource_name,
            "id": self.id,
            "backend_job_id": self.backend_job_id,
            "command": self.command,
            "image": self.image,
            "status": self.status.name,
            "num_nodes": self.num_nodes,
        }


class Machine:
    def __init__(
        self,
        name: str,
        backend_name: str,
        resource_name: Optional[str],
        image: str,
        status: JobStatus,
        backend_job_id: str,
    ):
        self.name = name
        self.backend_name = backend_name
        self.resource_name = resource_name
        self.image = image
        self.status = status
        self.backend_job_id = backend_job_id

    @classmethod
    def from_dict(cls, d):
        obj = cls(**d)
        obj.status = JobStatus(d["status"])
        return obj

    def to_dict(self):
        return {
            "name": self.name,
            "backend_name": self.backend_name,
            "resource_name": self.resource_name,
            "backend_job_id": self.backend_job_id,
            "image": self.image,
            "status": self.status.name,
        }


class Resource:
    def __init__(
        self,
        name: str,
        config: dict,
    ):
        self.name = name
        self.config = config

    @classmethod
    def from_dict(cls, d):
        obj = cls(**d)
        return obj

    def to_dict(self):
        d = {
            "name": self.name,
            "config": self.config,
        }
        return d


class Backend:
    def __init__(
        self,
        name: str,
        type: str,
        config: dict,
        resources: List[Resource],
    ):
        self.name = name
        self.type = type
        self.config = config
        self.resources = resources

    @classmethod
    def from_dict(cls, d):
        resources_dicts = d["resources"]
        resources = [Resource.from_dict(r) for r in resources_dicts]
        obj = cls(
            name=d["name"],
            type=d["type"],
            config=d["config"],
            resources=resources,
        )
        return obj

    def to_dict(self):
        d = {
            "name": self.name,
            "type": self.type,
            "config": self.config,
            "resources": [r.to_dict() for r in self.resources],
        }
        return d



import sys

import click
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.backends.backend import Backend
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.types import JobStatus
from komodo_cli.utils import APIClient, handle_errors, update_context_with_api_client


@click.command("shell")
@click.argument(
    "job_id",
    type=str,
)
@click.option(
    "--node-index",
    "-i",
    type=int,
    default=0,
    help="The index of the node to get logs from",
)
@click.pass_context
@handle_errors
def cmd_shell(ctx: click.Context, job_id: str, node_index: int):
    """Open a shell into a running Komodo job."""
    update_context_with_api_client(ctx)
    logger.info(f"Starting a shell into job {job_id}")

    api_client: APIClient = ctx.obj["api_client"]

    job = api_client.get_job(job_id)

    if job.status != JobStatus.RUNNING:
        printing.error(f"Job {job.id} is not running", bold=True)
        return

    backend_schema = api_client.get_backend(job.backend_name)
    backend: Backend = BackendFactory.get_backend(
        backend_schema,
        api_client,
    )

    printing.header(f"Shelling into job {job_id}", bold=True)
    backend.shell(job.backend_job_id, node_index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A CLI client for Komodo AI

.. currentmodule:: komodo_cli
.. moduleauthor:: kmushegi <mushegiani@gmail.com>
"""

from .version import __release__, __version__  # noqa
